library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity controlUnit is

  port (
    I_clock               : in  std_logic;  -- global clock
    I_reset               : in  std_logic;  -- asynchronous global reset
    I_inputSampleValid    : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_processingDone      : in  std_logic;

    O_loadShift_X         : out std_logic;  -- filtered sample
    O_loadShift_Y         : out std_logic;  -- Y sample
    O_loadShift_Z         : out std_logic;  -- Z sample
    
    O_initAddress         : out std_logic;  -- Control signal to initialize register read address
    O_incrAddress         : out std_logic;  -- Control signal to increment register read address
    O_X                   : out std_logic;
    O_Y                   : out std_logic;
    O_Z1                  : out std_logic;
    O_Z2                  : out std_logic;

    O_initSum             : out std_logic;  -- Control signal to initialize the MAC register
    O_loadSum             : out std_logic;  -- Control signal to load the MAC register;
    
    O_loadY               : out std_logic;  -- Control signal to load Y register
    
    O_FilteredSampleValid : out std_logic  -- Data valid signal for filtered sample
    );

end entity controlUnit;
architecture archi_operativeUnit of controlUnit is


  type T_state is (WAIT_SAMPLE, STORE_X, PROCESSING_LOOP_Y, STORE_Y, PROCESSING_LOOP_Z1, TRANSITION, PROCESSING_LOOP_Z2, STORE_Z, PROCESSING_LOOP_R, OUTPUT, WAIT_END_SAMPLE);  -- state list
  signal SR_presentState : T_state;
  signal SR_futurState   : T_state;

begin

  process (I_clock, I_reset) is
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
        SR_presentState <= WAIT_SAMPLE;
    elsif rising_edge(I_clock) then     -- rising clock edge
        SR_presentState <= SR_futurState;
    end if;
  end process;

  process (SR_presentState, I_inputSampleValid, I_processingDone) is
  begin
    case SR_presentState is

        when WAIT_SAMPLE =>
            if I_inputSampleValid = '1' then
                SR_futurState <= STORE_X;
            else
                SR_futurState <= WAIT_SAMPLE;
            end if;
        when STORE_X =>
            SR_futurState <= PROCESSING_LOOP_Y;

        when PROCESSING_LOOP_Y =>
            if I_processingDone = '1' then
                SR_futurState <= STORE_Y;
            else
                SR_futurState <= PROCESSING_LOOP_Y;
            end if;
        when STORE_Y =>
        SR_futurState <= PROCESSING_LOOP_Z1;

        when PROCESSING_LOOP_Z1 =>
            if I_processingDone = '1' then
                SR_futurState <= TRANSITION;
            else
                SR_futurState <= PROCESSING_LOOP_Z1;
            end if;
        when TRANSITION =>
        SR_futurState <= PROCESSING_LOOP_Z2;

        when PROCESSING_LOOP_Z2 =>
            if I_processingDone = '1' then
                SR_futurState <= STORE_Z;
            else
                SR_futurState <= PROCESSING_LOOP_Z2;
            end if;
        when STORE_Z =>
            SR_futurState <= PROCESSING_LOOP_R;
        
        when PROCESSING_LOOP_R =>
            if I_processingDone = '1' then
                SR_futurState <= OUTPUT;
            else
                SR_futurState <= PROCESSING_LOOP_R;
            end if;        
        when OUTPUT =>
            SR_futurState <= WAIT_END_SAMPLE;
        
        when WAIT_END_SAMPLE =>
            if I_inputSampleValid = '0' then
                SR_futurState <= WAIT_SAMPLE;
            else
                SR_futurState <= WAIT_END_SAMPLE;
            end if;
      when others => null;
    end case;
  end process;

    O_loadShift_X         <= '1' when SR_presentState = STORE_X else '0';
    O_loadShift_Y         <= '1' when SR_presentState = STORE_Y else '0';
    O_loadShift_Z         <= '1' when SR_presentState = STORE_Z else '0';
    O_initAddress         <= '1' when SR_presentState = STORE_X or SR_presentState = STORE_Y or SR_presentState = TRANSITION or SR_presentState = STORE_Z else '0';
    O_incrAddress         <= '1' when SR_presentState = PROCESSING_LOOP_Y or SR_presentState = PROCESSING_LOOP_Z1 or SR_presentState = PROCESSING_LOOP_Z2 or SR_presentState = PROCESSING_LOOP_R else '0';
    O_X                   <= '1' when SR_presentState = STORE_X else '0';
    O_Y                   <= '1' when SR_presentState = STORE_Y else '0';
    O_Z1                  <= '1' when SR_presentState = TRANSITION else '0';
    O_Z2                  <= '1' when SR_presentState = STORE_Z else '0';
    O_initSum             <= '1' when SR_presentState = STORE_X or SR_presentState = STORE_Y or SR_presentState = STORE_Z else '0';
    O_loadSum             <= '1' when SR_presentState = PROCESSING_LOOP_Y or SR_presentState = PROCESSING_LOOP_Z1 or SR_presentState = PROCESSING_LOOP_Z2 or SR_presentState = PROCESSING_LOOP_R else '0';
    O_loadY               <= '1' when SR_presentState = OUTPUT else '0';
    O_FilteredSampleValid <= '1' when SR_presentState = WAIT_END_SAMPLE else '0';


end architecture archi_operativeUnit;
